# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
MSS evaluation code on MUSDB18 test dataset.
'''

import test
import multiprocessing
import functools
import tqdm
import museval
import nnabla as nn
from nnabla.ext_utils import get_extension_context, import_extension_module
import musdb
from .args import get_inference_args


def separate_and_evaluate(track, args, ext):
    estimates = test.separate(track.audio, args)

    if args.out_dir:
        mus.save_estimates(estimates, track, args.out_dir)

    scores = museval.eval_mus_track(
        track, estimates, output_dir=args.out_dir
    )
    # clear cache memory
    ext.clear_memory_cache()
    return scores


if __name__ == '__main__':
    # Get the arguments parser
    _, args = get_inference_args()

    # Set NNabla context and Dynamic graph execution
    ctx = get_extension_context(args.context)
    nn.set_default_context(ctx)
    nn.set_auto_forward(True)
    ext = import_extension_module(args.context)

    mus = musdb.DB(
        root=args.root,
        download=args.root is None,
        subsets='test',
        is_wav=args.is_wav
    )

    if args.cores > 1:
        pool = multiprocessing.Pool(args.cores)
        results = museval.EvalStore()
        scores_list = list(
            pool.imap_unordered(
                func=functools.partial(separate_and_evaluate, args, ext),
                iterable=mus.tracks,
                chunksize=1
            )
        )
        pool.close()
        pool.join()
        for scores in scores_list:
            results.add_track(scores)

    else:
        results = museval.EvalStore()
        for track in tqdm.tqdm(mus.tracks):
            scores = separate_and_evaluate(track, args, ext)
            results.add_track(scores)

    print(results)
    method = museval.MethodStore()
    method.add_evalstore(results, args.out_dir)
    method.save(args.out_dir + '.pandas')



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
MSE and SDR Combination loss definition for MSS
'''

import nnabla.functions as F


def mse(x, y):
    # l2 distance and reduce mean
    se = F.squared_error(x, y)
    return F.mean(se)


def unsqueeze(x):
    # unsqueeze at first axis
    return F.reshape(x, (1,) + x.shape)


def mse_loss(mix_spec, msk_hat, gt_spec):
    # MSE-Combination Loss
    # mix_spec     -> (Fbin, BatchSize(16), 2(channels), Frames)
    # msk_hat      -> (Fbin, BatchSize(16), 8(4source x 2channels), Frames)
    #                       channel-dim -> [bass1, bass2, drums1, drums2, ...]
    # gt_spec      -> (Fbin, BatchSize(16), 8(4source x 2channels), Frames)

    assert msk_hat.shape == gt_spec.shape

    m_1 = msk_hat[Ellipsis, 0:2, :]  # bass
    m_2 = msk_hat[Ellipsis, 2:4, :]  # drums
    m_3 = msk_hat[Ellipsis, 4:6, :]  # vocals
    m_4 = msk_hat[Ellipsis, 6:8, :]  # other
    gt_1 = gt_spec[Ellipsis, 0:2, :]  # bass
    gt_2 = gt_spec[Ellipsis, 2:4, :]  # drums
    gt_3 = gt_spec[Ellipsis, 4:6, :]  # vocals
    gt_4 = gt_spec[Ellipsis, 6:8, :]  # other

    # 4C1 Combination Losses
    loss_mse_1 = mse(m_1*mix_spec, gt_1)
    loss_mse_2 = mse(m_2*mix_spec, gt_2)
    loss_mse_3 = mse(m_3*mix_spec, gt_3)
    loss_mse_4 = mse(m_4*mix_spec, gt_4)

    # 4C2 Combination Losses
    loss_mse_5 = mse((m_1+m_2)*mix_spec, (gt_1+gt_2))
    loss_mse_6 = mse((m_1+m_3)*mix_spec, (gt_1+gt_3))
    loss_mse_7 = mse((m_1+m_4)*mix_spec, (gt_1+gt_4))
    loss_mse_8 = mse((m_2+m_3)*mix_spec, (gt_2+gt_3))
    loss_mse_9 = mse((m_2+m_4)*mix_spec, (gt_2+gt_4))
    loss_mse_10 = mse((m_3+m_4)*mix_spec, (gt_3+gt_4))

    # 4C3 Combination Losses
    loss_mse_11 = mse((m_1+m_2+m_3)*mix_spec, (gt_1+gt_2+gt_3))
    loss_mse_12 = mse((m_1+m_2+m_4)*mix_spec, (gt_1+gt_2+gt_4))
    loss_mse_13 = mse((m_1+m_3+m_4)*mix_spec, (gt_1+gt_3+gt_4))
    loss_mse_14 = mse((m_2+m_3+m_4)*mix_spec, (gt_2+gt_3+gt_4))

    # All 14 Combination Losses (4C1 + 4C2 + 4C3)
    loss_mse = (loss_mse_1 + loss_mse_2 + loss_mse_3 + loss_mse_4 + loss_mse_5 + loss_mse_6 + loss_mse_7 +
                loss_mse_8 + loss_mse_9 + loss_mse_10 + loss_mse_11 + loss_mse_12 + loss_mse_13 + loss_mse_14) / 14.0

    return loss_mse


def sdr_loss(mix, pred, gt_time):
    # SDR-Combination Loss
    # mix     -> (BatchSize(16), 2(1 source x 2 channels), TimeLen)  -> (B, C, T)
    # pred    -> (4(sources), Bsize, 2(channels), Len)               -> (S, B, C, T)
    # gt_time -> (BatchSize(16), 8(4 source x 2 channels), TimeLen)  -> (B, S*C, T)
    #                       channel-dim -> [bass1, bass2, drums1, drums2, ...]

    _, batch_size, n_channels, length = pred.shape

    # Fix Length
    mix = mix[Ellipsis, :length]
    gt_time = gt_time[Ellipsis, :length]

    # Fix Shape
    mix = unsqueeze(mix)  # [1, B, C, T]
    gt_time = unsqueeze(gt_time)  # [1, B, S*C, T]
    data_t = mix  # [1, B, C, T]

    for i in range(4):
        data_t = F.concatenate(data_t, gt_time[Ellipsis, 2*i:2*i+2, :], axis=0)

    data_t = F.reshape(data_t, (-1, length))  # [5*B*C, T]
    pred = F.reshape(pred, (batch_size*n_channels *
                            pred.shape[0], pred.shape[-1]))  # [B*C*S, T]

    # Combination List (4C2 + 4C3)
    combi_list = [(1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4),
                  (1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4)]
    for combi in combi_list:
        if len(combi) == 2:
            tmp_data = data_t[batch_size*n_channels*combi[0]:batch_size*n_channels*(
                combi[0]+1), Ellipsis] + data_t[batch_size*n_channels*combi[1]:batch_size*n_channels*(combi[1]+1), Ellipsis]

            tmp_pred = pred[batch_size*n_channels*(combi[0]-1):batch_size*n_channels*combi[0], Ellipsis] + \
                pred[batch_size*n_channels*(
                                                   combi[1]-1):batch_size*n_channels*combi[1], Ellipsis]
        else:
            tmp_data = data_t[batch_size*n_channels*combi[0]:batch_size*n_channels*(combi[0]+1), Ellipsis] + data_t[batch_size*n_channels*combi[1]:batch_size*n_channels*(
                combi[1]+1), Ellipsis] + data_t[batch_size*n_channels*combi[2]:batch_size*n_channels*(combi[2]+1), Ellipsis]

            tmp_pred = pred[batch_size*n_channels*(combi[0]-1):batch_size*n_channels*combi[0], Ellipsis] + pred[batch_size*n_channels*(
                combi[1]-1):batch_size*n_channels*combi[1], Ellipsis] + pred[batch_size*n_channels*(combi[2]-1):batch_size*n_channels*combi[2], Ellipsis]

        data_t = F.concatenate(data_t, tmp_data, axis=0)
        pred = F.concatenate(pred, tmp_pred, axis=0)

    # All 14 Combinations (4C1 + 4C2 + 4C3)
    mix_t = F.tile(data_t[:batch_size*n_channels, Ellipsis], (14, 1))
    data_t = data_t[batch_size*n_channels:, Ellipsis]

    # SDR Loss Calculation
    loss_sdr = sdr_loss_core(pred, data_t, mix_t, weighted=True)

    return 1.0 + loss_sdr


def sdr_loss_core(inp, gt, mix, weighted=True):
    assert inp.shape == gt.shape  # (Batch, Len)
    assert mix.shape == gt.shape   # (Batch, Len)

    inp = inp[:, 200:-200]
    gt = gt[:, 200:-200]
    mix = mix[:, 200:-200]

    ns = mix - gt
    ns_hat = mix - inp

    if weighted:
        alpha = F.sum((gt*gt), 1, keepdims=True) / (F.sum((gt*gt), 1,
                                                          keepdims=True) + F.sum((ns*ns), 1, keepdims=True) + 1e-10)
    else:
        alpha = 0.5

    # Target
    num_cln = F.sum((inp*gt), 1, keepdims=True)
    denom_cln = ((1e-10 + F.sum((inp*inp), 1, keepdims=True))
                 ** 0.5) * ((1e-10 + F.sum((gt*gt), 1, keepdims=True)) ** 0.5)
    sdr_cln = num_cln / (denom_cln + 1e-10)

    # Noise
    num_noise = F.sum((ns*ns_hat), 1, keepdims=True)
    denom_noise = ((1e-10 + F.sum((ns_hat*ns_hat), 1, keepdims=True))
                   ** 0.5) * ((1e-10 + F.sum((ns*ns), 1, keepdims=True)) ** 0.5)
    sdr_noise = num_noise / (denom_noise + 1e-10)

    return F.mean(-alpha*sdr_cln - (1. - alpha)*sdr_noise)



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
ReduceLROnPlateau Learning Rate Scheduler code.
'''

import math


class ReduceLROnPlateau():
    """
    Reduce learning rate when loss has stopped reducing
    If loss does not reduce for 'patience' number of epochs, 
    learning rate is reduced by specified factor.
    Args:
        mode (str): either `min`, `max`. In `min` mode, lr will
            be reduced when the quantity monitored has stopped
            decreasing; in `max` mode it will be reduced when the
            quantity monitored has stopped increasing. 
            Default: 'min'.
        lr (float): learning rate before scheduler is put to work
            Default: 0
        factor (float): Factor by which the learning rate will be
            reduced. new_lr = lr * factor. 
            Default: 0.1.
        patience (int): Number of epochs with no improvement after
            which learning rate will be reduced. For example, if
            `patience = 2`, then we will ignore the first 2 epochs
            with no improvement, and will only decrease the LR after the
            3rd epoch if the loss still hasn't improved then.
            Default: 10.
        compare_mode (str): One of `rel`, `abs`. 
            In `rel` mode, current value must be '> best * ( 1 + delta)' or 
            '< best * ( 1 - delta)' to consider it is significant.
            In `abs` mode, current value must be '> best + delta' or 
            '< best - delta' to consider it is significant.
            Default: 'rel'
        cooldown (int): Number of epochs to wait before fiddling with learning 
            rate if it has already been reduced.
            Default: 0.
        min_lr (float): Lower bound on the learning rate. 
            Default: 0.
        eps (float): Minimal decay applied to lr. If the difference
            between new and old lr is smaller than eps, the update is
            ignored. 
            Default: 1e-8.
    Example:
        >>> scheduler = ReduceLROnPlateau('min')
        >>> for epoch in range(epochs):
        >>>     train(...)
        >>>     val_loss = validate(...)
        >>>     lr = scheduler.update_lr(val_loss, epoch)
    """

    def __init__(self, mode='min', lr=0, factor=0.1, patience=10,
                 compare_mode='rel', cooldown=10,
                 min_lr=0, eps=1e-8):

        if factor >= 1.0:
            raise ValueError('Factor should be < 1.0.')
        self.factor = factor
        self.lr = lr
        self._last_lr = None
        self.min_lr = min_lr
        self.patience = patience
        self.cooldown = cooldown
        self.cooldown_counter = 0
        self.mode = mode
        self.compare_mode = compare_mode
        self.best = None
        self.num_bad_epochs = None
        self.mode_worse = None  # the worse value for the chosen mode
        self.eps = eps
        self.last_epoch = 0
        self.delta = 1e-4
        self._init_is_better(mode=mode, compare_mode=compare_mode)
        self._reset()

    def _reset(self):
        """Resets num_bad_epochs counter and cooldown counter."""
        self.best = self.mode_worse
        self.cooldown_counter = 0
        self.num_bad_epochs = 0

    def update_lr(self, cur_metric, epoch=None):

        if epoch is None:
            epoch = self.last_epoch + 1
        self.last_epoch = epoch

        if self.is_better(cur_metric, self.best):
            self.best = cur_metric
            self.num_bad_epochs = 0
        else:
            self.num_bad_epochs += 1

        if self.in_cooldown():
            self.cooldown_counter -= 1
            self.num_bad_epochs = 0  # ignore any bad epochs in cooldown

        self._last_lr = self.lr

        if self.num_bad_epochs > self.patience:
            self._reduce_lr()
            self.cooldown_counter = self.cooldown
            self.num_bad_epochs = 0

        return self.lr

    def _reduce_lr(self):
        new_lr = max(self.lr * self.factor, self.min_lr)
        if self._last_lr - new_lr > self.eps:
            self._last_lr = self.lr
            self.lr = new_lr

    def in_cooldown(self):
        return self.cooldown_counter > 0

    def is_better(self, a, best):
        if self.mode == 'min' and self.compare_mode == 'rel':
            return a < best * (1. - self.delta)

        elif self.mode == 'min' and self.compare_mode == 'abs':
            return a < best - self.delta

        elif self.mode == 'max' and self.compare_mode == 'rel':
            return a > best * (self.delta + 1.)

        else:  # mode == 'max' and epsilon_mode == 'abs':
            return a > best + self.delta

    def _init_is_better(self, mode, compare_mode):
        if mode not in {'min', 'max'}:
            raise ValueError('mode ' + mode + ' is unknown!')
        if compare_mode not in {'rel', 'abs'}:
            raise ValueError('compare mode ' + compare_mode + ' is unknown!')

        if mode == 'min':
            self.mode_worse = math.inf
        else:  # mode == 'max':
            self.mode_worse = -(math.inf)

        self.mode = mode
        self.compare_mode = compare_mode



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
MUSDB18 data-iterator code for MSS.
'''

import random
import numpy as np
import musdb
from nnabla.utils.data_source import DataSource


class Compose():
    """Composes several augmentation transforms.
    Args:
        augmentations: list of augmentations to compose.
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, audio):
        for t in self.transforms:
            audio = t(audio)
        return audio


def _augment_gain(audio, low=0.25, high=1.25):
    """Applies a random gain between `low` and `high`"""
    g = random.uniform(low, high)
    return audio * g


def _augment_channelswap(audio):
    """Swap channels of stereo signals with a probability of p=0.5"""
    if audio.shape[0] == 2 and random.random() < 0.5:
        return np.flip(audio, 0)
    else:
        return audio


def load_datasources(parser, args):
    """Loads the specified dataset from commandline arguments
    Returns:
        train_dataset, validation_dataset
    """

    parser.add_argument('--is-wav', action='store_true', default=False,
                        help='loads wav instead of STEMS')
    parser.add_argument('--samples-per-track', type=int, default=64)
    parser.add_argument(
        '--source-augmentations', type=str, nargs='+',
        default=['gain', 'channelswap']
    )

    args = parser.parse_args()

    source_augmentations = Compose(
        [globals()['_augment_' + aug] for aug in args.source_augmentations]
    )

    train_dataset = MUSDBDataSource(
        source_augmentations=source_augmentations, random_track_mix=True, args=args)
    valid_dataset = MUSDBDataSource(
        split='valid', samples_per_track=1, args=args)

    return train_dataset, valid_dataset, args


class MUSDBDataSource(DataSource):
    """
    DataSource for MUSDB18 
    """

    def __init__(
        self,
        args,
        download=False,
        subsets='train',
        split='train',
        samples_per_track=64,
        source_augmentations=lambda audio: audio,
        random_track_mix=False,
        dtype=np.float32,
        seed=42,
        rng=None
    ):
        """
        MUSDB18 nnabla.utils.data_source that samples from the MUSDB tracks
        using track and excerpts with replacement.

        Parameters
        ----------
        args : additional arguments used to add further control for 
            the musdb dataset initialization function.
        download : boolean
            automatically download 7s preview version of MUS
        subsets : list-like [str]
            subset str or list of subset. Defaults to ``train``.
        split : str
            use (stratified) track splits for validation split (``valid``),
            defaults to ``train``.
        samples_per_track : int
            sets the number of samples, yielded from each track per epoch.
            Defaults to 64
        source_augmentations : list[callables]
            provide list of augmentation function that take a multi-channel
            audio file of shape (src, samples) as input and output. Defaults to
            no-augmentations (input = output)
        random_track_mix : boolean
            randomly mixes sources from different tracks to assemble a
            custom mix. This augmenation is only applied for the train subset.
        seed : int
            control randomness of dataset iterations
        dtype : numeric type
            data type of torch output tuple x and y
        """
        super(MUSDBDataSource, self).__init__(shuffle=(split == 'train'))
        if rng is None:
            rng = np.random.RandomState(seed)
        self.rng = rng

        random.seed(seed)
        self.args = args
        self.download = args.root is None
        self.subsets = subsets
        self.split = split
        self.samples_per_track = samples_per_track
        self.source_augmentations = source_augmentations
        self.random_track_mix = random_track_mix
        self.mus = musdb.DB(
            root=args.root,
            is_wav=args.is_wav,
            split=split,
            subsets=subsets,
            download=download
        )

        print(f"Finished loading dataset with {len(self.mus.tracks)} tracks.")

        self.sample_rate = 44100  # musdb is fixed sample rate
        self.dtype = dtype
        self._size = len(self.mus.tracks) * self.samples_per_track
        self._variables = ('mixture', 'target')
        self.reset()

    def _get_data(self, position):
        index = self._indexes[position]
        audio_sources = []

        # select track
        track = self.mus.tracks[index // self.samples_per_track]

        # at training time we assemble a custom mix
        if self.split == 'train' and self.args.seq_dur:
            for source in self.args.sources:
                # select a random track
                if self.random_track_mix:
                    track = random.choice(self.mus.tracks)

                # set the excerpt duration
                track.chunk_duration = self.args.seq_dur

                # set random start index
                track.chunk_start = random.uniform(
                    0, track.duration - self.args.seq_dur)

                # load source audio and apply time domain source_augmentations
                audio = track.sources[source].audio.T
                audio = self.source_augmentations(audio)
                audio_sources.append(audio)

            # create stem tensor of shape (source, channel, samples)
            stems = np.stack(audio_sources, axis=0)
            # apply linear mix over source index=0
            x = np.sum(stems, axis=0)
            if self.args.umx_train:
                # get the target stem for UMX training
                target_index = self.args.sources.index(self.args.source)
                y = stems[target_index]
            else:
                # get the target stems for X-UMX training
                y = stems[0]
                for idx in range(1, 4):
                    y = np.concatenate((y, stems[idx]), axis=0)

        # for validation and test, we deterministically yield the full musdb track
        else:
            # get the non-linear source mix straight from musdb
            x = track.audio.T
            if self.args.umx_train:
                # get the target stem for UMX validation
                y = track.targets[self.args.source].audio.T
            else:
                # get the target stems for X-UMX validation
                y = track.targets[self.args.sources[0]].audio.T
                for idx in range(1, 4):
                    tmp = track.targets[self.args.sources[idx]].audio.T
                    y = np.concatenate((y, tmp), axis=0)
        return x, y

    def reset(self):
        if self._shuffle:
            self._indexes = self.rng.permutation(self._size)
        else:
            self._indexes = np.arange(self._size)
        super(MUSDBDataSource, self).reset()



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
NNabla Utility Code for Communicators
'''

import nnabla as nn
from nnabla.logger import logger
from nnabla.ext_utils import get_extension_context
import nnabla.communicators as C
import logging


def create_float_context(ctx):
    ctx_float = get_extension_context(ctx.backend[0].split(':')[
                                      0], device_id=ctx.device_id)
    return ctx_float


class CommunicatorWrapper(object):
    def __init__(self, ctx):
        try:
            comm = C.MultiProcessDataParallelCommunicator(ctx)
            comm.init()
            self.n_procs = comm.size
            self.rank = comm.rank
            self.local_rank = comm.local_rank
            self.comm = comm
        except Exception as e:
            print(e)
            print('No communicator found. Running with a single process. If you run this with MPI processes,'
                  ' all processes will perform totally same.')
            self.n_procs = 1
            self.rank = 0
            self.local_rank = 0
            self.comm = None

        ctx.device_id = str(int(ctx.device_id) + int(self.local_rank))
        self.ctx = ctx
        self.ctx_float = create_float_context(ctx)

        logger.info("[Communicator] Using gpu_id = {} as rank = {}".format(
            self.ctx.device_id, self.rank))

    def all_reduce(self, params, division, inplace):
        if self.n_procs == 1:
            # skip all reduce since no processes have to be all-reduced
            return
        self.comm.all_reduce(params, division=division, inplace=inplace)

    def barrier(self):
        if self.n_procs == 1:
            return
        self.comm.barrier()

    def all_reduced_solver_update(self, solver, division=False, inplace=True):
        if self.n_procs > 1:
            params = [
                x.grad for x in solver.get_parameters().values()]
            self.all_reduce(params, division=division, inplace=inplace)

        solver.update()

    def all_reduced_solver_update_all(self, *solvers, division=False, inplace=True):
        for solver in solvers:
            self.all_reduced_solver_update(
                solver, division=division, inplace=inplace)

    def get_all_reduce_callback(self, packing_size=2 << 20):
        callback = None
        if self.n_procs > 1:
            params = [x.grad for x in nn.get_parameters().values()]
            callback = self.comm.all_reduce_callback(
                params, packing_size)
        return callback



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Utility code
'''

import re
import numpy as np
import sklearn.preprocessing
import tqdm
import nnabla as nn
from .model import get_stft, get_spectogram


def get_nnabla_version_integer():
    r = list(map(int, re.match('^(\d+)\.(\d+)\.(\d+)', nn.__version__).groups()))
    return r[0] * 10000 + r[1] * 100 + r[2]


def get_statistics(args, datasource):
    scaler = sklearn.preprocessing.StandardScaler()
    pbar = tqdm.tqdm(range(len(datasource.mus.tracks)))

    for ind in pbar:
        x = datasource.mus.tracks[ind].audio.T
        audio = nn.NdArray.from_numpy_array(x[None, ...])
        target_spec = get_spectogram(
            *get_stft(audio, n_fft=args.nfft, n_hop=args.nhop),
            mono=True
        )
        pbar.set_description("Compute dataset statistics")
        scaler.partial_fit(np.squeeze(target_spec.data))

    # set inital input scaler values
    std = np.maximum(
        scaler.scale_,
        1e-4*np.max(scaler.scale_)
    )
    return scaler.mean_, std


def bandwidth_to_max_bin(sample_rate, n_fft, bandwidth):
    freqs = np.linspace(
        0, float(sample_rate) / 2, n_fft // 2 + 1,
        endpoint=True
    )

    return np.max(np.where(freqs <= bandwidth)[0]) + 1


class AverageMeter():
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.sum += val * n
        self.count += n

    def get_avg(self):
        return self.sum / self.count


class EarlyStopping():
    def __init__(self, mode='min', min_delta=0, patience=10):
        self.mode = mode
        self.min_delta = min_delta
        self.patience = patience
        self.best = None
        self.num_bad_epochs = 0
        self.is_better = None
        self._init_is_better(mode, min_delta)

        if patience == 0:
            self.is_better = lambda a, b: True

    def step(self, metrics):
        if self.best is None:
            self.best = metrics
            return False

        if np.isnan(metrics):
            return True

        if self.is_better(metrics, self.best):
            self.num_bad_epochs = 0
            self.best = metrics
        else:
            self.num_bad_epochs += 1

        if self.num_bad_epochs >= self.patience:
            return True

        return False

    def _init_is_better(self, mode, min_delta):
        if mode not in {'min', 'max'}:
            raise ValueError('mode ' + mode + ' is unknown!')
        if mode == 'min':
            self.is_better = lambda a, best: a < best - min_delta
        if mode == 'max':
            self.is_better = lambda a, best: a > best + min_delta



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
ArgumentParser for X-UMX/UMX Inference/Trainer
'''

import argparse

def get_train_parser():
    parser = argparse.ArgumentParser(
        description='OpenUnmix_CrossNet(X-UMX)/OpenUnmix(UMX) Trainer')

    # Dataset paramaters
    parser.add_argument('--root', type=str,
                        help='root path of MUSDB18 dataset')
    parser.add_argument('--output', type=str, required=True,
                        help='provide output path base folder name for saving the checkpoints')

    # which target do we want to train?
    parser.add_argument('--source', type=str, default='vocals',
                        help='target source to be trained for UMX')
    parser.add_argument('--sources', type=str, nargs='+',
                        default=['bass', 'drums', 'vocals', 'other'],
                        help='List of target sources to be trained for X-UMX')
    parser.add_argument('--umx-train', action='store_true', default=False,
                        help='If True, OpenUnmix(UMX) network is used for Training')

    # Trainig Parameters
    parser.add_argument('--epochs', type=int, default=1000)
    parser.add_argument('--batch-size', type=int, default=16)
    parser.add_argument('--lr', type=float, default=0.001,
                        help='learning rate, defaults to 1e-3')
    parser.add_argument('--patience', type=int, default=140,
                        help='minimum number of bad epochs for EarlyStoping (default: 140 for UMX)')
    parser.add_argument('--lr-decay-patience', type=int, default=80,
                        help='lr decay patience for plateau scheduler')
    parser.add_argument('--lr-decay-gamma', type=float, default=0.3,
                        help='gamma of learning rate scheduler decay')
    parser.add_argument('--weight-decay', type=float, default=0.00001,
                        help='weight decay')
    parser.add_argument('--seed', type=int, default=42, metavar='S',
                        help='random seed (default: 42)')

    # Model Parameters
    parser.add_argument('--seq-dur', type=float, default=6.0,
                        help='Sequence duration in seconds per trainig batch'
                        'value of <=0.0 will use full/variable length')
    parser.add_argument('--unidirectional', action='store_true', default=False,
                        help='Use unidirectional LSTM instead of bidirectional')
    parser.add_argument('--nfft', type=int, default=4096,
                        help='STFT fft size and window size')
    parser.add_argument('--nhop', type=int, default=1024,
                        help='STFT hop size')
    parser.add_argument('--hidden-size', type=int, default=512,
                        help='hidden size parameter of dense bottleneck layers')
    parser.add_argument('--bandwidth', type=int, default=16000,
                        help='maximum model bandwidth in herz')
    parser.add_argument('--nb-channels', type=int, default=2,
                        help='set number of channels for model (1, 2)')
    parser.add_argument('--sample-rate', type=int, default=44100,
                        help='model sample rate')

    # Misc Parameters
    parser.add_argument('--mcoef', type=float, default=10.0,
                        help='coefficient for mixing: mfoef*TD-Loss + FD-Loss')
    parser.add_argument("--device-id", "-d", type=str, default='0',
                        help='Device ID the training run on. This is only valid if you specify `-c cudnn`.')
    parser.add_argument('--context', '-c', type=str,
                        default='cudnn', help="Extension modules. ex) 'cpu', 'cudnn'.")

    # The duration of validation sample
    parser.add_argument('--valid_dur', type=float, default=10.0,
                        help='Split duration for validation sample to avoid GPU memory overflow')
    return parser

def get_train_args():
    parser = get_train_parser()
    args, _ = parser.parse_known_args()
    return parser, args

def get_inference_parser():
    parser = argparse.ArgumentParser(
        description='OpenUnmix_CrossNet(X-UMX)/OpenUnmix(UMX) Inference/Evaluation')

    parser.add_argument('--inputs', type=str, nargs='+',
                        help='List of paths to any audio files supported by FFMPEG.')
    parser.add_argument('--targets', nargs='+', default=['bass', 'drums', 'vocals', 'other'],
                        type=str, help='provide targets to be processed. \
                        If none, all available targets will be computed')
    parser.add_argument('--out-dir', type=str,
                        help='Path to save separated sources')
    parser.add_argument('--start', type=float, default=0.0,
                        help='Audio chunk start in seconds')
    parser.add_argument('--duration', type=float, default=-1.0,
                        help='Audio chunk duration in seconds, negative values load full track')
    parser.add_argument('--model', type=str, required=True,
                        help='path to pretrained weights (weight filename in case of X-UMX Inference or directory of weight files for UMX Inference')
    parser.add_argument('--context', default='cudnn', type=str,
                        help='Execution on CUDA')
    parser.add_argument('--softmask', dest='softmask', action='store_true',
                        help=('if enabled, will initialize separation with softmask.'
                              'otherwise, will use mixture phase with spectrogram'))
    parser.add_argument('--niter', type=int, default=1,
                        help='number of iterations for refining results.')
    parser.add_argument('--alpha', type=float, default=1.0,
                        help='exponent in case of softmask separation')
    parser.add_argument('--sample-rate', type=int, default=44100,
                        help='model sample rate')
    parser.add_argument('--residual-model', action='store_true',
                        help='create a model for the residual')
    parser.add_argument('--chunk-dur', type=int, default=None,
                        help='window length in seconds - reduce this if Inference fails with SegFault')
    parser.add_argument('--umx-infer', action='store_true', default=False,
                        help='If True, OpenUnmix(UMX) network is used for Inference/Evaluation')

    # Only Evaluation specific arguments
    parser.add_argument('--root', type=str,
                        help='root path of MUSDB18 dataset')
    parser.add_argument('--cores', type=int, default=1)
    parser.add_argument('--is-wav', action='store_true',
                        default=False, help='flags wav version of the dataset')
    return parser

def get_inference_args():
    parser = get_inference_parser()
    args, _ = parser.parse_known_args()
    return parser, args



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
MSS Inference code using X-UMX/UMX.
'''

from pathlib import Path
import os
import warnings
from tqdm import trange
import norbert
import scipy.signal
import resampy
import soundfile as sf
import numpy as np
import nnabla as nn
from nnabla.ext_utils import get_extension_context
from pydub import AudioSegment
from pydub.utils import mediainfo
from .args import get_inference_args
from .model import *
from .utils import bandwidth_to_max_bin


def istft(X, rate=44100, n_fft=4096, n_hopsize=1024):
    _, audio = scipy.signal.istft(
        X / (n_fft // 2),
        rate,
        nperseg=n_fft,
        noverlap=n_fft - n_hopsize,
        boundary=True
    )
    return audio


def separate(audio, args):
    """
    Performing the separation on audio input
    Parameters
    ----------
    audio: np.ndarray [shape=(nb_timesteps, nb_channels)]
        mixture audio
    args : ArgumentParser
        ArgumentParser for OpenUnmix_CrossNet(X-UMX)/OpenUnmix(UMX) Inference

    Returns
    -------
    estimates: `dict` [`str`, `np.ndarray`]
        dictionary of all estimates as performed by the separation model.
    """

    # convert numpy audio to NNabla Variable
    audio_nn = nn.Variable.from_numpy_array(audio.T[None, ...])
    source_names = []
    V = []
    max_bin = bandwidth_to_max_bin(
        sample_rate=44100, n_fft=4096, bandwidth=16000)

    if not args.umx_infer:
        # Run X-UMX Inference
        nn.load_parameters(args.model)
        for j, target in enumerate(args.targets):
            if j == 0:
                unmix_target = OpenUnmix_CrossNet(
                    max_bin=max_bin, is_predict=True)
                mix_spec, msk, _ = unmix_target(audio_nn, test=True)
                # Network output is (nb_frames, nb_samples, nb_channels, nb_bins)
            V.append((msk[Ellipsis, j * 2:j * 2 + 2, :]
                      * mix_spec).d[:, 0, ...])
            source_names += [target]
    else:
        # Run UMX Inference
        for j, target in enumerate(args.targets):
            with nn.parameter_scope(target):
                unmix_target = OpenUnmix(max_bin=max_bin)
                nn.load_parameters(f"{os.path.join(args.model, target)}.h5")
                # Network output is (nb_frames, nb_samples, nb_channels, nb_bins)
                V.append(unmix_target(audio_nn, test=True).d[:, 0, ...])
            source_names += [target]

    V = np.transpose(np.array(V), (1, 3, 2, 0))
    if args.softmask:
        # only exponentiate the model if we use softmask
        V = V ** args.alpha

    real, imag = get_stft(audio_nn, center=True)

    # convert to complex numpy type
    X = real.d + imag.d * 1j
    X = X[0].transpose(2, 1, 0)

    if args.residual_model or len(args.targets) == 1:
        V = norbert.residual_model(V, X, args.alpha if args.softmask else 1)
        source_names += (['residual'] if len(args.targets) > 1
                         else ['accompaniment'])

    Y = norbert.wiener(V, X.astype(np.complex128), args.niter,
                       use_softmask=args.softmask)

    estimates = {}
    for j, name in enumerate(source_names):
        audio_hat = istft(
            Y[..., j].T,
            n_fft=unmix_target.n_fft,
            n_hopsize=unmix_target.n_hop
        )
        estimates[name] = audio_hat.T

    return estimates


def test():
    _, args = get_inference_args()

    # Set NNabla context and Dynamic graph execution
    ctx = get_extension_context(args.context)
    nn.set_default_context(ctx)

    # Enable the NNabla Dynamic excecution
    nn.set_auto_forward(True)

    for input_file in args.inputs:

        # get audio data from the path - all formats recognized by FFMPEG are recognized
        audio_with_meta = AudioSegment.from_file(input_file)
        sample_rate = int(mediainfo(input_file)['sample_rate'])
        channel_sounds = audio_with_meta.split_to_mono()
        samples = [s.get_array_of_samples()
                   for idx, s in enumerate(channel_sounds) if idx < 2]
        fp_arr = np.array(samples).T.astype(np.float32)
        fp_arr /= np.iinfo(samples[0].typecode).max
        audio = fp_arr

        if audio.shape[1] > 2:
            # if it is multi-channel audio, consider only first two channels
            warnings.warn(
                'Channel count > 2! '
                'Only the first two channels will be processed!')
            audio = audio[:, :2]

        if sample_rate != args.sample_rate:
            # resample to model samplerate if needed
            audio = resampy.resample(
                audio, sample_rate, args.sample_rate, axis=0)

        if audio.shape[1] == 1:
            # if we have mono, let's duplicate it
            # as the input of OpenUnmix is always stereo
            audio = np.repeat(audio, 2, axis=1)

        if args.chunk_dur is not None:
            # split and separate sources using moving window protocol for each chunk of audio
            # chunk duration must be lower for machines with low memory
            chunk_size = sample_rate * args.chunk_dur
            if (audio.shape[0] % chunk_size) == 0:
                nchunks = (audio.shape[0] // chunk_size)
            else:
                nchunks = (audio.shape[0] // chunk_size) + 1
        else:
            chunk_size = audio.shape[0]
            nchunks = 1

        estimates = {}
        for chunk_idx in trange(nchunks):
            cur_chunk = audio[chunk_idx *
                              chunk_size: min((chunk_idx + 1) * chunk_size, audio.shape[0]), :]
            cur_estimates = separate(cur_chunk, args)
            if any(estimates) is False:
                estimates = cur_estimates
            else:
                for key in cur_estimates:
                    estimates[key] = np.concatenate(
                        (estimates[key], cur_estimates[key]), axis=0)

        if not args.out_dir:
            model_path = Path(args.model)
            if not model_path.exists():
                output_path = Path(Path(input_file).stem + '_' + model_path.stem)
            else:
                output_path = Path(
                    Path(input_file).stem + '_' + model_path.stem
                )
        else:
            if len(args.inputs) > 1:
                output_path = Path(args.out_dir) / Path(input_file).stem
            else:
                output_path = Path(args.out_dir)

        output_path.mkdir(exist_ok=True, parents=True)

        for target, estimate in estimates.items():
            sf.write(
                str(output_path / Path(target).with_suffix('.wav')),
                estimate,
                args.sample_rate
            )


if __name__ == '__main__':
    test()



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
MSS Training code using X-UMX/UMX.
'''

import os
import nnabla as nn
import nnabla.solvers as S
from nnabla.ext_utils import get_extension_context, import_extension_module
from nnabla.utils.data_iterator import data_iterator
from nnabla.monitor import Monitor, MonitorSeries, MonitorTimeElapsed
from numpy.random import RandomState, seed
from tqdm import trange
from .lr_scheduler import ReduceLROnPlateau
from .comm import CommunicatorWrapper
from .args import get_train_args
from .model import get_model
from .data import load_datasources
from .utils import *
seed(42)


def train():
    # Check NNabla version
    if get_nnabla_version_integer() < 11900:
        raise ValueError(
            'Please update the nnabla version to v1.19.0 or latest version since memory efficiency of core engine is improved in v1.19.0')

    parser, args = get_train_args()

    # Get context.
    ctx = get_extension_context(args.context, device_id=args.device_id)
    comm = CommunicatorWrapper(ctx)
    nn.set_default_context(comm.ctx)
    ext = import_extension_module(args.context)

    # Monitors
    # setting up monitors for logging
    monitor_path = args.output
    monitor = Monitor(monitor_path)

    monitor_best_epoch = MonitorSeries(
        'Best epoch', monitor, interval=1)
    monitor_traing_loss = MonitorSeries(
        'Training loss', monitor, interval=1)
    monitor_validation_loss = MonitorSeries(
        'Validation loss', monitor, interval=1)
    monitor_lr = MonitorSeries(
        'learning rate', monitor, interval=1)
    monitor_time = MonitorTimeElapsed(
        "training time per iteration", monitor, interval=1)

    if comm.rank == 0:
        if not os.path.isdir(args.output):
            os.makedirs(args.output)

    # Initialize DataIterator for MUSDB18.
    train_source, valid_source, args = load_datasources(parser, args)

    train_iter = data_iterator(
        train_source,
        args.batch_size,
        RandomState(args.seed),
        with_memory_cache=False,
    )

    valid_iter = data_iterator(
        valid_source,
        1,
        RandomState(args.seed),
        with_memory_cache=False,
    )

    if comm.n_procs > 1:
        train_iter = train_iter.slice(
            rng=None, num_of_slices=comm.n_procs, slice_pos=comm.rank)

        valid_iter = valid_iter.slice(
            rng=None, num_of_slices=comm.n_procs, slice_pos=comm.rank)

    # Calculate maxiter per GPU device.
    # Change max_iter, learning_rate and weight_decay according no. of gpu devices for multi-gpu training.
    default_batch_size = 16
    train_scale_factor = (comm.n_procs * args.batch_size) / default_batch_size
    max_iter = int((train_source._size // args.batch_size) // comm.n_procs)
    weight_decay = args.weight_decay * train_scale_factor
    args.lr = args.lr * train_scale_factor

    # Calculate the statistics (mean and variance) of the dataset
    scaler_mean, scaler_std = get_statistics(args, train_source)

    # clear cache memory
    ext.clear_memory_cache()

    max_bin = bandwidth_to_max_bin(
        train_source.sample_rate, args.nfft, args.bandwidth
    )

    # Get X-UMX/UMX computation graph and variables as namedtuple
    model = get_model(args, scaler_mean, scaler_std, max_bin=max_bin)

    # Create Solver and set parameters.
    solver = S.Adam(args.lr)
    solver.set_parameters(nn.get_parameters())

    # Initialize Early Stopping
    es = EarlyStopping(patience=args.patience)

    # Initialize LR Scheduler (ReduceLROnPlateau)
    lr_scheduler = ReduceLROnPlateau(
        lr=args.lr, factor=args.lr_decay_gamma, patience=args.lr_decay_patience)
    best_epoch = 0

    # AverageMeter for mean loss calculation over the epoch
    losses = AverageMeter()

    # Training loop.
    for epoch in trange(args.epochs):
        # TRAINING
        losses.reset()
        for batch in range(max_iter):
            model.mixture_audio.d, model.target_audio.d = train_iter.next()
            solver.zero_grad()
            model.loss.forward(clear_no_need_grad=True)
            if comm.n_procs > 1:
                all_reduce_callback = comm.get_all_reduce_callback()
                model.loss.backward(clear_buffer=True,
                                    communicator_callbacks=all_reduce_callback)
            else:
                model.loss.backward(clear_buffer=True)
            solver.weight_decay(weight_decay)
            solver.update()
            losses.update(model.loss.d.copy(), args.batch_size)
        training_loss = losses.get_avg()

        # clear cache memory
        ext.clear_memory_cache()

        # VALIDATION
        losses.reset()
        for batch in range(int(valid_source._size // comm.n_procs)):
            x, y = valid_iter.next()
            dur = int(valid_source.sample_rate * args.valid_dur)
            sp, cnt = 0, 0
            loss_tmp = nn.NdArray()
            loss_tmp.zero()
            while 1:
                model.vmixture_audio.d = x[Ellipsis, sp:sp+dur]
                model.vtarget_audio.d = y[Ellipsis, sp:sp+dur]
                model.vloss.forward(clear_no_need_grad=True)
                cnt += 1
                sp += dur
                loss_tmp += model.vloss.data
                if x[Ellipsis, sp:sp+dur].shape[-1] < dur or x.shape[-1] == cnt*dur:
                    break
            loss_tmp = loss_tmp / cnt
            if comm.n_procs > 1:
                comm.all_reduce(loss_tmp, division=True, inplace=True)
            losses.update(loss_tmp.data.copy(), 1)
        validation_loss = losses.get_avg()

        # clear cache memory
        ext.clear_memory_cache()

        lr = lr_scheduler.update_lr(validation_loss, epoch=epoch)
        solver.set_learning_rate(lr)
        stop = es.step(validation_loss)

        if comm.rank == 0:
            monitor_best_epoch.add(epoch, best_epoch)
            monitor_traing_loss.add(epoch, training_loss)
            monitor_validation_loss.add(epoch, validation_loss)
            monitor_lr.add(epoch, lr)
            monitor_time.add(epoch)

            if validation_loss == es.best:
                best_epoch = epoch
                # save best model
                if args.umx_train:
                    nn.save_parameters(os.path.join(
                        args.output, 'best_umx.h5'))
                else:
                    nn.save_parameters(os.path.join(
                        args.output, 'best_xumx.h5'))

        if args.umx_train:
            # Early stopping for UMX after `args.patience` (140) number of epochs
            if stop:
                print("Apply Early Stopping")
                break


if __name__ == '__main__':
    train()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="xumx-spleeterweb",
    version="0.2.3",
    author="JeffreyCA",
    author_email="jeffreyca16@gmail.com",
    description="Unofficial NNabla implementation of CrossNet-Open-Unmix (X-UMX), originally created by Sony Research AI.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JeffreyCA/xumx",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'lameenc>=1.2.2',
        'musdb>=0.3.1',
        'museval>=0.3.0',
        'requests>=2.22',
        'scipy>=1.3.1',
        'setuptools>=41.0.0',
        'norbert>=0.2.1',
        'resampy>=0.2.2',
        'nnabla>=1.13.0',
        'pydub>=0.24.1',
        'scikit-learn>=0.23.2'
    ],
    python_requires='>=3.6',
)



# Copyright (c) 2021 Sony Corporation. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
X-UMX/UMX Architecture definition for MSS.
'''

import collections
import nnabla as nn
from nnabla.parameter import get_parameter_or_create
import nnabla.functions as F
import nnabla.parametric_functions as PF
import numpy as np
from .loss import mse_loss, sdr_loss


def get_stft(x, n_fft=4096, n_hop=1024, center=True):
    '''
    Multichannel STFT

    Input: (nb_samples, nb_channels, nb_timesteps)
    Output: (nb_samples, nb_channels, nb_bins, nb_frames), 
            (nb_samples, nb_channels, nb_bins, nb_frames)
    '''
    nb_samples, nb_channels, _ = x.shape
    x = F.reshape(x, (nb_samples*nb_channels, -1))

    real, imag = F.stft(
        x, n_fft, n_hop, n_fft,
        window_type='hanning',
        center=center,
        pad_mode='reflect'
    )

    real = F.reshape(real,
                     (nb_samples, nb_channels, n_fft // 2 + 1, -1)
                     )

    imag = F.reshape(imag,
                     (nb_samples, nb_channels, n_fft // 2 + 1, -1)
                     )

    return real, imag


def get_spectogram(real, imag, power=1, mono=True):
    '''
    Input:  (nb_samples, nb_channels, nb_bins, nb_frames), 
            (nb_samples, nb_channels, nb_bins, nb_frames)
    Output: (nb_frames, nb_samples, nb_channels, nb_bins)
    '''
    spec = ((real ** 2) + (imag ** 2)) ** (power / 2.0)

    if mono:
        spec = F.mean(spec, axis=1, keepdims=True)

    return F.transpose(spec, ((3, 0, 1, 2)))


class BaseClass():
    def __init__(
        self,
        test=False,
        n_fft=4096,
        n_hop=1024,
        input_is_spectrogram=False,
        hidden_size=512,
        nb_channels=2,
        nb_layers=3,
        nb_of_directions=2,
        input_mean=None,
        input_scale=None,
        max_bin=None,
    ):
        self.test = test
        self.nb_output_bins = n_fft // 2 + 1
        self.input_is_spectrogram = input_is_spectrogram
        self.hidden_size = hidden_size
        self.n_fft = n_fft
        self.n_hop = n_hop
        self.nb_channels = nb_channels
        self.nb_layers = nb_layers
        self.nb_of_directions = nb_of_directions
        self.input_mean = input_mean
        self.input_scale = input_scale

        if max_bin:
            self.nb_bins = max_bin
        else:
            self.nb_bins = self.nb_output_bins

        if input_mean is None:
            self.input_mean = np.zeros((self.nb_bins,))
        if input_scale is None:
            self.input_scale = np.ones((self.nb_bins,))

    def create_input_mean_parameters(self, name):
        param = get_parameter_or_create(name, shape=(
            self.nb_bins,), initializer=-self.input_mean[:self.nb_bins])
        return F.reshape(param, (1, 1, 1, self.nb_bins))

    def create_input_scale_parameters(self, name):
        param = get_parameter_or_create(name, shape=(
            self.nb_bins,), initializer=1.0/self.input_scale[:self.nb_bins])
        return F.reshape(param, (1, 1, 1, self.nb_bins))

    def create_output_parameters(self, name):
        param = get_parameter_or_create(name, shape=(
            self.nb_output_bins,), initializer=np.ones((self.nb_output_bins,)))
        return F.reshape(param, (1, 1, 1, self.nb_output_bins))

    def lstm(self, lstm_in, nb_samples, scope_name):
        '''
        Apply 3-layered LSTM
        '''
        h = F.constant(shape=(self.nb_layers, self.nb_of_directions,
                              nb_samples, self.hidden_size // 2))
        c = F.constant(shape=(self.nb_layers, self.nb_of_directions,
                              nb_samples, self.hidden_size // 2))
        lstm_out, _, _ = PF.lstm(lstm_in, h, c, num_layers=self.nb_layers,
                                 bidirectional=True, training=not self.test, dropout=0.4, name=scope_name)
        return lstm_out

    def fc_bn(self, fc_bn_in, out_channels, scope_name, activation=None):
        '''
        Apply dense and batch norm layer
        '''
        fc_out = PF.affine(fc_bn_in, out_channels, base_axis=2,
                           with_bias=False, name=scope_name)
        fc_bn_out = PF.batch_normalization(
            fc_out, axes=[2], batch_stat=not self.test, name=scope_name)
        if activation == 'tanh':
            fc_bn_out = F.tanh(fc_bn_out)
        elif activation == 'relu':
            fc_bn_out = F.relu(fc_bn_out)
        return fc_bn_out


class OpenUnmix_CrossNet(BaseClass):
    def __init__(
        self,
        input_mean=None,
        input_scale=None,
        max_bin=None,
        is_predict=False
    ):
        super(OpenUnmix_CrossNet, self).__init__(
            input_mean=input_mean, input_scale=input_scale, max_bin=max_bin)

        self.is_predict = is_predict

        # Initilize input mean for all the 4 categories of track
        self.input_mean_bass = self.create_input_mean_parameters(
            'input_mean_bass')
        self.input_mean_drums = self.create_input_mean_parameters(
            'input_mean_drums')
        self.input_mean_vocals = self.create_input_mean_parameters(
            'input_mean_vocals')
        self.input_mean_other = self.create_input_mean_parameters(
            'input_mean_other')

        # Initilize input scale for all the 4 categories of track
        self.input_scale_bass = self.create_input_scale_parameters(
            'input_scale_bass')
        self.input_scale_drums = self.create_input_scale_parameters(
            'input_scale_drums')
        self.input_scale_vocals = self.create_input_scale_parameters(
            'input_scale_vocals')
        self.input_scale_other = self.create_input_scale_parameters(
            'input_scale_other')

        # Initilize output scale for all the 4 categories of track
        self.output_scale_bass = self.create_output_parameters(
            'output_scale_bass')
        self.output_scale_drums = self.create_output_parameters(
            'output_scale_drums')
        self.output_scale_vocals = self.create_output_parameters(
            'output_scale_vocals')
        self.output_scale_other = self.create_output_parameters(
            'output_scale_other')

        # Initilize output mean for all the 4 categories of track
        self.output_mean_bass = self.create_output_parameters(
            'output_mean_bass')
        self.output_mean_drums = self.create_output_parameters(
            'output_mean_drums')
        self.output_mean_vocals = self.create_output_parameters(
            'output_mean_vocals')
        self.output_mean_other = self.create_output_parameters(
            'output_mean_other')

    def __call__(self, x, test=False):
        '''
        Input: (nb_samples, nb_channels, nb_timesteps)
            or (nb_frames, nb_samples, nb_channels, nb_bins)
        Outputs: Input Power/Mag Spectrogram, Output Power/Mag Spectrogram and Predictd sources
        '''
        self.test = test
        fft_real, fft_imag = get_stft(x, n_fft=self.n_fft, n_hop=self.n_hop)
        x_theta = F.atan2(fft_imag, fft_real)
        x = get_spectogram(fft_real, fft_imag, mono=(self.nb_channels == 1))

        nb_frames, nb_samples, nb_channels, nb_bins = x.shape

        mix_spec = F.identity(x)
        x = x[..., :self.nb_bins]

        # clone
        x_bass = F.identity(x)
        x_drums = F.identity(x)
        x_vocals = F.identity(x)
        x_other = F.identity(x)

        # shift and scale input to mean=0 std=1 (across all bins)
        x_bass += self.input_mean_bass
        x_drums += self.input_mean_drums
        x_vocals += self.input_mean_vocals
        x_other += self.input_mean_other

        x_bass *= self.input_scale_bass
        x_drums *= self.input_scale_drums
        x_vocals *= self.input_scale_vocals
        x_other *= self.input_scale_other

        # encode and normalize every instance in a batch
        x_bass = self.fc_bn(x_bass, self.hidden_size,
                            "fc1_bass", activation='tanh')
        x_drums = self.fc_bn(x_drums, self.hidden_size,
                             "fc1_drums", activation='tanh')
        x_vocals = self.fc_bn(x_vocals, self.hidden_size,
                              "fc1_vocals", activation='tanh')
        x_other = self.fc_bn(x_other, self.hidden_size,
                             "fc1_other", activation='tanh')

        # Average the sources
        cross_1 = (x_bass + x_drums + x_vocals + x_other) / 4.0

        # apply 3-layers of stacked LSTM
        lstm_out_bass = self.lstm(cross_1, nb_samples, "lstm_bass")
        lstm_out_drums = self.lstm(cross_1, nb_samples, "lstm_drums")
        lstm_out_vocals = self.lstm(cross_1, nb_samples, "lstm_vocals")
        lstm_out_other = self.lstm(cross_1, nb_samples, "lstm_other")

        # lstm skip connection
        x_bass = F.concatenate(x_bass, lstm_out_bass)
        x_drums = F.concatenate(x_drums, lstm_out_drums)
        x_vocals = F.concatenate(x_vocals, lstm_out_vocals)
        x_other = F.concatenate(x_other, lstm_out_other)

        cross_2 = (x_bass + x_drums + x_vocals + x_other) / 4.0

        # first dense stage + batch norm
        x_bass = self.fc_bn(cross_2, self.hidden_size,
                            "fc2_bass", activation='relu')
        x_drums = self.fc_bn(cross_2, self.hidden_size,
                             "fc2_drums", activation='relu')
        x_vocals = self.fc_bn(cross_2, self.hidden_size,
                              "fc2_vocals", activation='relu')
        x_other = self.fc_bn(cross_2, self.hidden_size,
                             "fc2_other", activation='relu')

        # second dense stage + batch norm
        x_bass = self.fc_bn(x_bass, nb_channels*nb_bins, "fc3_bass")
        x_drums = self.fc_bn(x_drums, nb_channels*nb_bins, "fc3_drums")
        x_vocals = self.fc_bn(x_vocals, nb_channels*nb_bins, "fc3_vocals")
        x_other = self.fc_bn(x_other, nb_channels*nb_bins, "fc3_other")

        # reshape back to original dim
        x_bass = F.reshape(x_bass, (nb_frames, nb_samples,
                                    nb_channels, self.nb_output_bins))
        x_drums = F.reshape(x_drums, (nb_frames, nb_samples,
                                      nb_channels, self.nb_output_bins))
        x_vocals = F.reshape(
            x_vocals, (nb_frames, nb_samples, nb_channels, self.nb_output_bins))
        x_other = F.reshape(x_other, (nb_frames, nb_samples,
                                      nb_channels, self.nb_output_bins))

        # apply output scale and shift
        x_bass *= self.output_scale_bass
        x_drums *= self.output_scale_drums
        x_vocals *= self.output_scale_vocals
        x_other *= self.output_scale_other

        x_bass += self.output_mean_bass
        x_drums += self.output_mean_drums
        x_vocals += self.output_mean_vocals
        x_other += self.output_mean_other

        # since our output is non-negative, we can apply RELU
        mask_bass = F.relu(x_bass)
        mask_drums = F.relu(x_drums)
        mask_vocals = F.relu(x_vocals)
        mask_other = F.relu(x_other)

        # (Frames, Bsize, Channels, Fbins)
        x_bass = mask_bass * mix_spec
        x_drums = mask_drums * mix_spec
        x_vocals = mask_vocals * mix_spec
        x_other = mask_other * mix_spec

        if not self.is_predict:
            tmp = F.stack(*[x_bass, x_drums, x_vocals, x_other], axis=0)
            # (4(sources), Frames, Bsize(16), 2(channels), Fbins) ==> (4, Bsize, Channels, Fbins, Frames)
            tmp = F.transpose(tmp, (0, 2, 3, 4, 1))
            pred_r, pred_i = [], []
            for i in range(tmp.shape[0]):
                pred_r.append(tmp[i] * F.cos(x_theta))
                pred_i.append(tmp[i] * F.sin(x_theta))
            pred_r = F.stack(*pred_r, axis=0)
            pred_i = F.stack(*pred_i, axis=0)
            pred_r = F.reshape(
                pred_r, (4*nb_samples*nb_channels, self.nb_output_bins, nb_frames))
            pred_i = F.reshape(
                pred_i, (4*nb_samples*nb_channels, self.nb_output_bins, nb_frames))
            pred = F.istft(pred_r, pred_i, self.n_fft, self.n_hop,
                           self.n_fft, window_type='hanning', center=True)
            pred = F.reshape(pred, (4, nb_samples, nb_channels, -1))
        else:
            pred = None

        return mix_spec, F.concatenate(mask_bass, mask_drums, mask_vocals, mask_other, axis=2), pred


class OpenUnmix(BaseClass):
    def __init__(
        self,
        input_mean=None,
        input_scale=None,
        max_bin=None,
    ):
        super(OpenUnmix, self).__init__(input_mean=input_mean,
                                        input_scale=input_scale, max_bin=max_bin)

        self.input_mean = self.create_input_mean_parameters('input_mean')
        self.input_scale = self.create_input_scale_parameters('input_scale')
        self.output_scale = self.create_output_parameters('output_scale')
        self.output_mean = self.create_output_parameters('output_mean')

    def __call__(self, x, test=False):
        '''
        Input: (nb_samples, nb_channels, nb_timesteps)
            or (nb_frames, nb_samples, nb_channels, nb_bins)
        Outputs: Output Power/Mag Spectrogram
        '''
        self.test = test
        if not self.input_is_spectrogram:
            x = get_spectogram(*get_stft(x, n_fft=self.n_fft, n_hop=self.n_hop,
                                         center=self.test), mono=(self.nb_channels == 1))
        nb_frames, nb_samples, nb_channels, nb_bins = x.shape
        mix_spec = F.identity(x)

        # crop
        x = x[..., :self.nb_bins]

        # shift and scale input to mean=0 std=1 (across all bins)
        x += self.input_mean
        x *= self.input_scale

        # encode and normalize every instance in a batch
        x = self.fc_bn(x, self.hidden_size, "fc1", activation='tanh')

        # apply 3-layers of stacked LSTM
        lstm_out = self.lstm(x, nb_samples, "lstm")

        # lstm skip connection
        x = F.concatenate(x, lstm_out)

        # first dense stage + batch norm
        x = self.fc_bn(x, self.hidden_size, "fc2", activation='relu')

        # second dense stage + batch norm
        x = self.fc_bn(x, nb_channels*nb_bins, "fc3")

        # reshape back to original dim
        x = F.reshape(x, (nb_frames, nb_samples,
                          nb_channels, self.nb_output_bins))

        # apply output scaling
        x *= self.output_scale
        x += self.output_mean

        return F.relu(x) * mix_spec


def get_model(args, input_mean, input_scale, max_bin=None):
    '''
    Create computation graph and variables for X-UMX/UMX.
    '''
    # target channels (2 for UMX and 8 for X-UMX (2 * 4-target sources
    target_channels = args.nb_channels if args.umx_train else 4 * args.nb_channels

    # Create input variables.
    mixture_audio = nn.Variable(
        (args.batch_size, args.nb_channels, args.sample_rate * args.seq_dur))
    target_audio = nn.Variable(
        (args.batch_size, target_channels, args.sample_rate * args.seq_dur))
    vmixture_audio = nn.Variable(
        (1, args.nb_channels, args.sample_rate * args.valid_dur))
    vtarget_audio = nn.Variable(
        (1, target_channels, args.sample_rate * args.valid_dur))

    if args.umx_train:
        # create training graph for UMX
        unmix = OpenUnmix(input_mean=input_mean,
                          input_scale=input_scale, max_bin=max_bin)
        pred_spec = unmix(mixture_audio)
        target_spec = get_spectogram(*get_stft(target_audio, n_fft=unmix.n_fft,
                                               n_hop=unmix.n_hop, center=False), mono=(unmix.nb_channels == 1))
        loss = F.mean(F.squared_error(pred_spec, target_spec))

        # create validation graph for UMX
        vpred_spec = unmix(vmixture_audio, test=True)
        vtarget_spec = get_spectogram(
            *get_stft(vtarget_audio, n_fft=unmix.n_fft, n_hop=unmix.n_hop), mono=(unmix.nb_channels == 1))
        vloss = F.mean(F.squared_error(vpred_spec, vtarget_spec))
    else:
        # create training graph for X-UMX
        unmix = OpenUnmix_CrossNet(
            input_mean=input_mean, input_scale=input_scale, max_bin=max_bin)
        mix_spec, m_hat, pred = unmix(mixture_audio)
        target_spec = get_spectogram(
            *get_stft(target_audio, n_fft=unmix.n_fft, n_hop=unmix.n_hop), mono=(unmix.nb_channels == 1))
        loss_f = mse_loss(mix_spec, m_hat, target_spec)
        loss_t = sdr_loss(mixture_audio, pred, target_audio)
        loss = args.mcoef * loss_t + loss_f

        # create validation graph for X-UMX
        vmix_spec, vm_hat, vpred = unmix(vmixture_audio, test=True)
        vtarget_spec = get_spectogram(
            *get_stft(vtarget_audio, n_fft=unmix.n_fft, n_hop=unmix.n_hop), mono=(unmix.nb_channels == 1))
        vloss_f = mse_loss(vmix_spec, vm_hat, vtarget_spec)
        vloss_t = sdr_loss(vmixture_audio, vpred, vtarget_audio)
        vloss = args.mcoef * vloss_t + vloss_f

    loss.persistent = True
    vloss.persistent = True

    Network = collections.namedtuple(
        'Network', 'loss, vloss, mixture_audio, target_audio, vmixture_audio, vtarget_audio')
    return Network(
        loss=loss,
        vloss=vloss,
        mixture_audio=mixture_audio,
        target_audio=target_audio,
        vmixture_audio=vmixture_audio,
        vtarget_audio=vtarget_audio
    )
